// Copyright (C) 2015  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_CONTRoL_
#define DLIB_CONTRoL_

#include "control/lspi.h"
#include "control/mpc.h"

#endif // DLIB_CONTRoL_


