// Copyright (C) 2003  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_SERVER_IOSTREAM_CPp_
#define DLIB_SERVER_IOSTREAM_CPp_

#include "server_iostream.h"

namespace dlib
{
    const logger server_iostream::_dLog("dlib.server_iostream");
}

#endif // DLIB_SERVER_IOSTREAM_CPp_

