// Copyright (C) 2005  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_GUI_CORe_
#define DLIB_GUI_CORe_


#include "platform.h"



#ifdef WIN32
#include "gui_core/windows.h"
#else
#include "gui_core/xlib.h"
#endif



#endif // DLIB_GUI_CORe_

