// Copyright (C) 2013  Davis E. King (davis@dlib.net)
// License: Boost Software License   See LICENSE.txt for the full license.
#ifndef DLIB_PYTHoN_TOP_
#define DLIB_PYTHoN_TOP_ 

#include "python/boost_python_utils.h"
#include "python/pyassert.h"
#include "python/serialize_pickle.h"
#include "python/numpy.h"
#include "python/numpy_image.h"

#endif // DLIB_PYTHoN_TOP_


